-------------------------------- MODULE 1Bit --------------------------------
\* 1Bit implementation of mutual exclusion
EXTENDS Integers

CONSTANT N
ASSUME N \in Nat

Procs == 0..N



(********
--algorithm 1BitProtocol {
    variables flag = [i \in Procs |-> FALSE], turn = 0 ;
    fair process (P \in Procs) {
     ncs:- while (TRUE) {
            skip ;
     enter: flag[self] := TRUE ;
        e2: if (flag[1-self]) {
        e3:   if (self /= turn) {
        e4:     flag[self] := FALSE;
                goto e3;
              }
              else {
                goto enter;
              }
            };
        cs: skip ;
      exit: flag[self] := FALSE ;
        x2: turn := 1-self;
    }
  }
}
********)
\* BEGIN TRANSLATION (chksum(pcal) = "9162e634" /\ chksum(tla) = "3b0cd6ea")
VARIABLES flag, turn, pc

vars == << flag, turn, pc >>

ProcSet == (Procs)

Init == (* Global variables *)
        /\ flag = [i \in Procs |-> FALSE]
        /\ turn = 0
        /\ pc = [self \in ProcSet |-> "ncs"]

ncs(self) == /\ pc[self] = "ncs"
             /\ TRUE
             /\ pc' = [pc EXCEPT ![self] = "enter"]
             /\ UNCHANGED << flag, turn >>

enter(self) == /\ pc[self] = "enter"
               /\ flag' = [flag EXCEPT ![self] = TRUE]
               /\ pc' = [pc EXCEPT ![self] = "e2"]
               /\ turn' = turn

e2(self) == /\ pc[self] = "e2"
            /\ IF flag[1-self]
                  THEN /\ pc' = [pc EXCEPT ![self] = "e3"]
                  ELSE /\ pc' = [pc EXCEPT ![self] = "cs"]
            /\ UNCHANGED << flag, turn >>

e3(self) == /\ pc[self] = "e3"
            /\ IF self /= turn
                  THEN /\ pc' = [pc EXCEPT ![self] = "e4"]
                  ELSE /\ pc' = [pc EXCEPT ![self] = "enter"]
            /\ UNCHANGED << flag, turn >>

e4(self) == /\ pc[self] = "e4"
            /\ flag' = [flag EXCEPT ![self] = FALSE]
            /\ pc' = [pc EXCEPT ![self] = "e3"]
            /\ turn' = turn

cs(self) == /\ pc[self] = "cs"
            /\ TRUE
            /\ pc' = [pc EXCEPT ![self] = "exit"]
            /\ UNCHANGED << flag, turn >>

exit(self) == /\ pc[self] = "exit"
              /\ flag' = [flag EXCEPT ![self] = FALSE]
              /\ pc' = [pc EXCEPT ![self] = "x2"]
              /\ turn' = turn

x2(self) == /\ pc[self] = "x2"
            /\ turn' = 1-self
            /\ pc' = [pc EXCEPT ![self] = "ncs"]
            /\ flag' = flag

P(self) == ncs(self) \/ enter(self) \/ e2(self) \/ e3(self) \/ e4(self)
              \/ cs(self) \/ exit(self) \/ x2(self)

Next == (\E self \in Procs: P(self))

Spec == /\ Init /\ [][Next]_vars
        /\ \A self \in Procs : WF_vars((pc[self] # "ncs") /\ P(self))

\* END TRANSLATION 

--------

\* Invariant which implies MutualExclusion
\*Inv == \A i \in Procs : (pc[i] = "cs") => (flag[i] /\ (~flag[1-i] \/ pc[1-i] = "e2"))
\*Inv == \A i \in Procs : (pc[i] = "cs") => (flag[i] /\ (~flag[1-i] \/ pc[1-i] \in {"e2","e3"}))
Inv == \A i \in Procs : (pc[i] = "cs") => (flag[i] /\ (~flag[1-i] \/ pc[1-i] \in {"e2","e3","e4"}))


ME == INSTANCE MutualExclusionSpec WITH
                pc <- [proc \in Procs  |->  CASE pc[proc] = "cs"             ->  "cs"
                                              [] pc[proc] \in {"enter","e2","e3","e4","e5"} ->  "csentry"
                                              [] OTHER                       ->  "non-cs"]

Live0 == (pc[0] = "enter") ~> (pc[0] = "cs")
Live1 == (pc[1] = "enter") ~> (pc[1] = "cs")

THEOREM Spec => ME!Spec
THEOREM Spec => []ME!MutualExclusion
THEOREM Spec => ME!Liveness


=============================================================================
