----------------------------- MODULE Alternate -----------------------------
\* Module Alternate specifies turn-based solution for mutual-exclusion problem.
EXTENDS Integers

CONSTANT Procs
\*ASSUME \E N \in Nat : Procs = 0..N
ASSUME \E N \in 0..10 : Procs = 0..N

Max(s) == CHOOSE x \in s :
                    /\ \A y \in s: y =< x
                    /\ \E z \in s: z = x

NProcs == Max(Procs)+1 \* XXX assumes procs are 0..N


(********
--algorithm Alternate {
  variable turn \in Procs;
  fair process (p \in Procs) {
    ncs: while (TRUE) {
            skip;   \* non-critical section

  enter:    await turn = self;

     cs:    skip;   \* critical section

\*   exit:    turn := 1 - self
   exit:    turn := ((self + 1) % NProcs)
         };
  }
}
********)
\* BEGIN TRANSLATION (chksum(pcal) = "5d5d2a74" /\ chksum(tla) = "579f6ceb")
VARIABLES turn, pc

vars == << turn, pc >>

ProcSet == (Procs)

Init == (* Global variables *)
        /\ turn \in Procs
        /\ pc = [self \in ProcSet |-> "ncs"]

ncs(self) == /\ pc[self] = "ncs"
             /\ TRUE
             /\ pc' = [pc EXCEPT ![self] = "enter"]
             /\ turn' = turn

enter(self) == /\ pc[self] = "enter"
               /\ turn = self
               /\ pc' = [pc EXCEPT ![self] = "cs"]
               /\ turn' = turn

cs(self) == /\ pc[self] = "cs"
            /\ TRUE
            /\ pc' = [pc EXCEPT ![self] = "exit"]
            /\ turn' = turn

exit(self) == /\ pc[self] = "exit"
              /\ turn' = ((self + 1) % NProcs)
              /\ pc' = [pc EXCEPT ![self] = "ncs"]

p(self) == ncs(self) \/ enter(self) \/ cs(self) \/ exit(self)

Next == (\E self \in Procs: p(self))

Spec == /\ Init /\ [][Next]_vars
        /\ \A self \in Procs : WF_vars(p(self))

\* END TRANSLATION 


--------



ME == INSTANCE MutualExclusionSpec WITH
                pc <- [proc \in Procs  |->  CASE pc[proc] = "cs"     ->  "cs"
                                              [] pc[proc] = "enter"  ->  "csentry"
                                              [] OTHER               ->  "non-cs"]

THEOREM Spec => ME!Spec
THEOREM Spec => []ME!MutualExclusion
THEOREM Spec => ME!Liveness


KeepsEnteringCS == \A i \in Procs : []<>(pc[i] = "cs")
Stuck           == \E i \in Procs : <>[](pc[i] = "ncs")


=============================================================================
